/**
* The MIT License (MIT)
* 
* Copyright (c) 2013 Retroswitch LLC
* 
* Permission is hereby granted, free of charge, to any person obtaining a copy
* of this software and associated documentation files (the "Software"), to deal
* in the Software without restriction, including without limitation the rights
* to use, copy, modify, merge, publish, distribute, sublicense, and/or sell
* copies of the Software, and to permit persons to whom the Software is
* furnished to do so, subject to the following conditions:
* 
* The above copyright notice and this permission notice shall be included in
* all copies or substantial portions of the Software.
* 
* THE SOFTWARE IS PROVIDED "AS IS", WITHOUT WARRANTY OF ANY KIND, EXPRESS OR
* IMPLIED, INCLUDING BUT NOT LIMITED TO THE WARRANTIES OF MERCHANTABILITY,
* FITNESS FOR A PARTICULAR PURPOSE AND NONINFRINGEMENT. IN NO EVENT SHALL THE
* AUTHORS OR COPYRIGHT HOLDERS BE LIABLE FOR ANY CLAIM, DAMAGES OR OTHER
* LIABILITY, WHETHER IN AN ACTION OF CONTRACT, TORT OR OTHERWISE, ARISING FROM,
* OUT OF OR IN CONNECTION WITH THE SOFTWARE OR THE USE OR OTHER DEALINGS IN
* THE SOFTWARE.
*
*/

//
// Includes

#include "Canvas.h"
#include "ProfileManager.h"
#include <QMouseEvent>
#include <QPainter>

//
// Canvas Methods

/******************************************************************************/
/* Constructor                                                                */
/******************************************************************************/

Canvas::Canvas( QWidget* parent ) :
    QWidget( parent ), m_profile( 0 ), m_drawMode( cDrawModeNone ), m_borderColor( 0 ),
    m_backgroundColor( 0 ), m_drawingColor( 0 ), m_display( 0 )
{
}

/******************************************************************************/
/* Destructor                                                                 */
/******************************************************************************/

Canvas::~Canvas()
{
    delete m_display;
}

/******************************************************************************/
/* SetProfile                                                                 */
/******************************************************************************/

void Canvas::SetProfile( const Profile* profile, void* initialDisplayContents )
{
    if( m_profile != profile )
    {
        m_profile = profile;

        delete m_display;
        m_display = 0;

        m_borderColor = 0;
        m_backgroundColor = 0;
        m_drawingColor = 0;

        if( m_profile )
        {
            m_display = new uint8[m_profile->m_rows * m_profile->m_columns];

            if( initialDisplayContents )
            {
                memcpy( m_display, initialDisplayContents, m_profile->m_rows * m_profile->m_columns );
            }
            else
            {
                memset( m_display, 0xff, m_profile->m_rows * m_profile->m_columns );
            }
        }

        RecomputeDisplayRect();
        repaint();
    }
}

/******************************************************************************/
/* Clear                                                                      */
/******************************************************************************/

void Canvas::Clear()
{
    if( m_profile )
    {
        memset( m_display, 0xff, m_profile->m_rows * m_profile->m_columns );
        repaint();
    }
}

/******************************************************************************/
/* Plot                                                                       */
/******************************************************************************/

void Canvas::Plot( uint8 x, uint8 y, uint8 color )
{
    if( m_profile )
    {
        m_display[y * m_profile->m_columns + x] = color;
        repaint();
    }
}

/******************************************************************************/
/* paintEvent                                                                 */
/******************************************************************************/

void Canvas::paintEvent( QPaintEvent* event )
{
    (void)event;

    QPainter painter( this );

    if( m_profile )
    {
        // Generate background/border display
    
        painter.fillRect( rect(), m_profile->m_colors.at( m_borderColor ) );
        painter.fillRect( m_displayRect, m_profile->m_colors.at( m_backgroundColor ) );

        // Plot characters / cells

        float cell_width = m_displayRect.width() / m_profile->m_columns;
        float cell_height = m_displayRect.height() / m_profile->m_rows;
        float cell_height_scale = m_profile->m_scanlineGap ? 7.f / 8.f : 1.f;

        for( int y=0; y<m_profile->m_rows; y++ )
        {
            for( int x=0; x<m_profile->m_columns; x++ )
            {
                uint8 color = m_display[y * m_profile->m_columns + x];

                if( color < m_profile->m_colors.count() )
                {
                    QRectF cell( m_displayRect.x() + x * cell_width, m_displayRect.y() + y * cell_height, cell_width, cell_height * cell_height_scale );
                    painter.fillRect( cell, m_profile->m_colors.at( color ) );
                }
            }
        }
    }
    else
    {
        painter.fillRect( rect(), QColor( "#444" ) );
    }
}

/******************************************************************************/
/* resizeEvent                                                                */
/******************************************************************************/

void Canvas::resizeEvent( QResizeEvent* event )
{
    (void)event;

    RecomputeDisplayRect();
    repaint();
}

/******************************************************************************/
/* mousePressEvent                                                            */
/******************************************************************************/

void Canvas::mousePressEvent( QMouseEvent* event )
{
    if( m_profile )
    {
        switch( event->button() )
        {
            case Qt::LeftButton:
                m_drawMode = cDrawModeDraw;
                Plot( event->pos() );
                break;

            case Qt::RightButton:
                m_drawMode = cDrawModeErase;
                Plot( event->pos() );
                break;

            default:
                break;
        }
    }
}

/******************************************************************************/
/* mouseReleaseEvent                                                          */
/******************************************************************************/

void Canvas::mouseReleaseEvent( QMouseEvent* event )
{
    (void)event;

    m_drawMode = cDrawModeNone;
}

/******************************************************************************/
/* mouseMoveEvent                                                             */
/******************************************************************************/

void Canvas::mouseMoveEvent( QMouseEvent* event )
{
    if( m_drawMode != cDrawModeNone )
    {
        Plot( event->pos() );
    }
}

/******************************************************************************/
/* RecomputeDisplayRect                                                       */
/******************************************************************************/

void Canvas::RecomputeDisplayRect()
{
    if( m_profile )
    {
        float client_aspect = rect().width() / float( rect().height() );
        float display_aspect = m_profile->m_columns / float( m_profile->m_rows ) * m_profile->m_pixelAspect;

        float cell_width;
        float cell_height;

        if( display_aspect < client_aspect )
        {
            // Bound by height
            float min_border_thickness = rect().height() * .1f;
            cell_height = (rect().height() - min_border_thickness * 2) / float( m_profile->m_rows );
            cell_width = cell_height * m_profile->m_pixelAspect;
        }
        else
        {
            // Bound by width
            float min_border_thickness = rect().width() * .1f;
            cell_width = (rect().width() - min_border_thickness * 2) / float( m_profile->m_columns );
            cell_height = cell_width / m_profile->m_pixelAspect;
        }

        float display_width = cell_width * m_profile->m_columns;
        float display_height = cell_height * m_profile->m_rows;

        m_displayRect.setRect(
            rect().left() + (rect().width() * .5f) - (display_width * .5f),
            rect().top() + (rect().height() * .5f) - (display_height * .5f),
            display_width, display_height );
    }
}

/******************************************************************************/
/* Plot                                                                       */
/******************************************************************************/

void Canvas::Plot( const QPoint& mousePos )
{
    float cell_width = m_displayRect.width() / m_profile->m_columns;
    float cell_height = m_displayRect.height() / m_profile->m_rows;

    int cell_x = int( (mousePos.x() - m_displayRect.left()) / cell_width );
    int cell_y = int( (mousePos.y() - m_displayRect.top()) / cell_height );

    if( cell_x < 0 || cell_x >= m_profile->m_columns )
    {
        return;
    }

    if( cell_y < 0 || cell_y >= m_profile->m_rows )
    {
        return;
    }

    uint8& color = m_display[cell_y * m_profile->m_columns + cell_x];
    uint8 pen_color = (m_drawMode == cDrawModeDraw) ? m_drawingColor : 0xff;

    if( color != pen_color )
    {
        color = pen_color;
        repaint();

        emit CellChanged( uint8( cell_x ), uint8( cell_y ), pen_color );
    }
}
