/**
* The MIT License (MIT)
* 
* Copyright (c) 2013 Retroswitch LLC
* 
* Permission is hereby granted, free of charge, to any person obtaining a copy
* of this software and associated documentation files (the "Software"), to deal
* in the Software without restriction, including without limitation the rights
* to use, copy, modify, merge, publish, distribute, sublicense, and/or sell
* copies of the Software, and to permit persons to whom the Software is
* furnished to do so, subject to the following conditions:
* 
* The above copyright notice and this permission notice shall be included in
* all copies or substantial portions of the Software.
* 
* THE SOFTWARE IS PROVIDED "AS IS", WITHOUT WARRANTY OF ANY KIND, EXPRESS OR
* IMPLIED, INCLUDING BUT NOT LIMITED TO THE WARRANTIES OF MERCHANTABILITY,
* FITNESS FOR A PARTICULAR PURPOSE AND NONINFRINGEMENT. IN NO EVENT SHALL THE
* AUTHORS OR COPYRIGHT HOLDERS BE LIABLE FOR ANY CLAIM, DAMAGES OR OTHER
* LIABILITY, WHETHER IN AN ACTION OF CONTRACT, TORT OR OTHERWISE, ARISING FROM,
* OUT OF OR IN CONNECTION WITH THE SOFTWARE OR THE USE OR OTHER DEALINGS IN
* THE SOFTWARE.
*
*/

#ifndef _CANVAS_H
#define _CANVAS_H

//
// Includes

#include "Types.h"
#include <QWidget>

//
// Forward Declarations

class Profile;
class QPaintEvent;
class QResizeEvent;
class QMouseEvent;

//
// Canvas Class

class Canvas : public QWidget
{
    Q_OBJECT

public:

    // --- Construction/Initialization ---

                            Canvas( QWidget* parent = 0 );
    virtual                ~Canvas();

    // --- Configuration ---

                            // Get/set current machine profile.
    const Profile*          GetProfile() const;
    void                    SetProfile( const Profile* profile, void* initialDisplayContents = 0 );

                            // Get/set current border color.
    uint8                   GetBorderColor() const;
    void                    SetBorderColor( uint8 colorIndex );

                            // Get/set current background color.
    uint8                   GetBackgroundColor() const;
    void                    SetBackgroundColor( uint8 colorIndex );

                            // Get/set current drawing color.
    uint8                   GetDrawingColor() const;
    void                    SetDrawingColor( uint8 colorIndex );

    // --- Drawing ---

                            // Clear display memory.
    void                    Clear();

                            // Plot a single color/cell. If color is 0xff, the
                            // cell will be erased.
    void                    Plot( uint8 x, uint8 y, uint8 color );

signals:

                            // This signal is emitted whenever a cell is changed
                            // as a result of a drawing operation.
    void                    CellChanged( uint8 x, uint8 y, uint8 color );

private:

    // --- Private Types ---

    enum DrawMode
    {
        cDrawModeNone = -1,

        cDrawModeDraw,
        cDrawModeErase,

        cNumDrawModes
    };

    // --- QWidget Overrides ---

    virtual void            paintEvent( QPaintEvent* event );
    virtual void            resizeEvent( QResizeEvent* event );
    virtual void            mousePressEvent( QMouseEvent* event );
    virtual void            mouseReleaseEvent( QMouseEvent* event );
    virtual void            mouseMoveEvent( QMouseEvent* event );

    // --- Private Methods ---

    void                    RecomputeDisplayRect();
    void                    Plot( const QPoint& mousePos );

    // --- Private Members ---

    const Profile*          m_profile;
    QRectF                  m_displayRect;

    DrawMode                m_drawMode;

    uint8                   m_borderColor;
    uint8                   m_backgroundColor;
    uint8                   m_drawingColor;

    uint8*                  m_display;

};

//
// Canvas Inline Methods

/******************************************************************************/
/* GetProfile                                                                 */
/******************************************************************************/

inline const Profile* Canvas::GetProfile() const
{
    return( m_profile );
}

/******************************************************************************/
/* GetBorderColor                                                             */
/******************************************************************************/

inline uint8 Canvas::GetBorderColor() const
{
    return( m_borderColor );
}

/******************************************************************************/
/* SetBorderColor                                                             */
/******************************************************************************/

inline void Canvas::SetBorderColor( uint8 colorIndex )
{
    m_borderColor = colorIndex;
    repaint();
}

/******************************************************************************/
/* GetBackgroundColor                                                         */
/******************************************************************************/

inline uint8 Canvas::GetBackgroundColor() const
{
    return( m_backgroundColor );
}

/******************************************************************************/
/* SetBackgroundColor                                                         */
/******************************************************************************/

inline void Canvas::SetBackgroundColor( uint8 colorIndex )
{
    m_backgroundColor = colorIndex;
    repaint();
}

/******************************************************************************/
/* GetDrawingColor                                                            */
/******************************************************************************/

inline uint8 Canvas::GetDrawingColor() const
{
    return( m_drawingColor );
}

/******************************************************************************/
/* SetDrawingColor                                                            */
/******************************************************************************/

inline void Canvas::SetDrawingColor( uint8 colorIndex )
{
    m_drawingColor = colorIndex;
}

//
//

#endif // _CANVAS_H