/**
* The MIT License (MIT)
* 
* Copyright (c) 2013 Retroswitch LLC
* 
* Permission is hereby granted, free of charge, to any person obtaining a copy
* of this software and associated documentation files (the "Software"), to deal
* in the Software without restriction, including without limitation the rights
* to use, copy, modify, merge, publish, distribute, sublicense, and/or sell
* copies of the Software, and to permit persons to whom the Software is
* furnished to do so, subject to the following conditions:
* 
* The above copyright notice and this permission notice shall be included in
* all copies or substantial portions of the Software.
* 
* THE SOFTWARE IS PROVIDED "AS IS", WITHOUT WARRANTY OF ANY KIND, EXPRESS OR
* IMPLIED, INCLUDING BUT NOT LIMITED TO THE WARRANTIES OF MERCHANTABILITY,
* FITNESS FOR A PARTICULAR PURPOSE AND NONINFRINGEMENT. IN NO EVENT SHALL THE
* AUTHORS OR COPYRIGHT HOLDERS BE LIABLE FOR ANY CLAIM, DAMAGES OR OTHER
* LIABILITY, WHETHER IN AN ACTION OF CONTRACT, TORT OR OTHERWISE, ARISING FROM,
* OUT OF OR IN CONNECTION WITH THE SOFTWARE OR THE USE OR OTHER DEALINGS IN
* THE SOFTWARE.
*
*/

//
// Includes

#include "PaintServer.h"
#include "Flyer.h"
#include "Display.h"
#include "Message.h"
#include <conio.h>
#include <stdio.h>
#include <stdlib.h>

//
// Type Definitions

typedef enum tagState
{
    cStateDisconnected,
    cStateListening,
    cStateConnected,
    
    cNumStates
} State;

//
// Defines & Macros

#define PETSCII_STOP  0x03
#define NETWORK_CHANNEL 2

//
// Local Prototypes

void localHandleConnectMessage( MessageHeader* header );
void localHandleCommandMessage( MessageHeader* header );
void localHandleScanlineMessage( MessageHeader* header );

//
// Functions

/******************************************************************************/
/* PaintServerStartup                                                         */
/******************************************************************************/

int PaintServerStartup()
{
    // Open command channel, verify the Flyer is there...

    if( FlyerStartup() < 0 )
    {
        cprintf( "No response from Flyer.\r\n" );
        return( -1 );
    }

    // Make sure firmware supports the features we'll be using

    if( FlyerRequireFirmware( "1.2.1" ) != 1 )
    {
        cprintf( "Flyer firmware 1.2.1 or later is required.\r\n" );
        FlyerShutdown();
        return( -1 );
    }

    DisplayStartup();

    return( 0 );
}

/******************************************************************************/
/* PaintServerShutdown                                                        */
/******************************************************************************/

void PaintServerShutdown()
{
    DisplayShutdown();
    FlyerShutdown();
}

/******************************************************************************/
/* PaintServerRun                                                             */
/******************************************************************************/

void PaintServerRun()
{
    State state = cStateDisconnected;
    uint8 running = 1;

    while( running )
    {
        switch( state )
        {
            case cStateDisconnected:
            {
                uint8 ip[4];
                char buf[32];
                FlyerGetIpAddress( ip );

                sprintf( buf, "%d.%d.%d.%d:%d", ip[0], ip[1], ip[2], ip[3], PAINTSERVER_PORT );
                DisplayStatusEnable( "Listening (STOP exits)", buf );

                FlyerListen( NETWORK_CHANNEL, PAINTSERVER_PORT );
                state = cStateListening;

                break;
            }
            
            case cStateListening:
            {
                if( FlyerGetBytesAvailable( NETWORK_CHANNEL ) >= 0 )
                {
                    DisplayStatusDisable();
                    state = cStateConnected;
                }

                break;
            }

            case cStateConnected:
            {
                MessageHeader header;
                int8 status = MessageGetHeader( NETWORK_CHANNEL, &header );

                if( status < 0 )
                {
                    cbm_close( NETWORK_CHANNEL );
                    state = cStateDisconnected;
                }
                else if( status )
                {
                    switch( header.m_id )
                    {
                        case cMessageIdConnect:
                            localHandleConnectMessage( &header );
                            break;

                        case cMessageIdCommand:
                            localHandleCommandMessage( &header );
                            break;

                        case cMessageIdScanline:
                            localHandleScanlineMessage( &header );
                            break;
                    }
                }

                break;
            }
        }

        if( kbhit() && cgetc() == PETSCII_STOP )
        {
            running = 0;
        }
    }

    if( state != cStateDisconnected )
    {
        cbm_close( NETWORK_CHANNEL );
    }
}

//
// Local Functions

/******************************************************************************/
/* localHandleConnectMessage                                                  */
/******************************************************************************/

void localHandleConnectMessage( MessageHeader* header )
{
    ConnectPayload payload;
    MessageHeader update_header;

    cbm_read( NETWORK_CHANNEL, &payload, header->m_payloadLength );

    // Verify server / client protocol matches and respond with canvas update

    if( payload.m_protocol == MESSAGE_PROTOCOL )
    {
        UpdatePayload update_payload;
        uint8 tmp[80];
        uint8 x;
        uint8 y;

        update_header.m_id = cMessageIdUpdate;
        update_header.m_pad = 0;
        update_header.m_payloadLength = sizeof( UpdatePayload ) + DisplayGetWidth() * DisplayGetHeight();

#if defined(__PET__)
        update_payload.m_target = (DisplayGetWidth() == 40) ? 0 : 1;
#elif defined(__C64__)
        update_payload.m_target = 3;
#endif
        update_payload.m_borderColor = DisplayGetBorderColor();
        update_payload.m_backgroundColor = DisplayGetBackgroundColor();

        cbm_write( NETWORK_CHANNEL, &update_header, sizeof( MessageHeader ) );
        cbm_write( NETWORK_CHANNEL, &update_payload, sizeof( UpdatePayload ) );

        for( y=0; y<DisplayGetHeight(); y++ )
        {
            for( x=0; x<DisplayGetWidth(); x++ )
            {
                tmp[x] = DisplayGetCell( x, y );
            }

            cbm_write( NETWORK_CHANNEL, tmp, DisplayGetWidth() );
        }
    }
    else
    {
        cbm_close( NETWORK_CHANNEL );
    }
}

/******************************************************************************/
/* localHandleCommandMessage                                                  */
/******************************************************************************/

void localHandleCommandMessage( MessageHeader* header )
{
    CommandPayload payload;
    cbm_read( NETWORK_CHANNEL, &payload, header->m_payloadLength );
    
    switch( payload.m_id )
    {
        case cCommandIdClear:
            DisplayClear();
            break;

        case cCommandIdSetColors:
            DisplaySetBorderColor( payload.m_arg[0] );
            DisplaySetBackgroundColor( payload.m_arg[1] );
            break;

        case cCommandIdPlot:
            DisplaySetCell( payload.m_arg[0], payload.m_arg[1], payload.m_arg[2] );
            break;
    }
}

/******************************************************************************/
/* localHandleScanlineMessage                                                 */
/******************************************************************************/

void localHandleScanlineMessage( MessageHeader* header )
{
    uint8 buf[80 + sizeof( ScanlinePayload )];
    ScanlinePayload* payload = (ScanlinePayload*)buf;
    uint8 x;
    uint8* src = &buf[sizeof( ScanlinePayload )];

    cbm_read( NETWORK_CHANNEL, payload, header->m_payloadLength );

    for( x=0; x<payload->m_length; x++ )
    {
        DisplaySetCell( x, payload->m_row, *src++ );
    }
}
